"""Module for Azure-related functionality."""

import json
from pathlib import Path

from azure.identity import DefaultAzureCredential
from azure.mgmt.authorization import AuthorizationManagementClient
from azure.mgmt.containerregistry import ContainerRegistryManagementClient
from azure.mgmt.containerservice import ContainerServiceClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.resource import ResourceManagementClient, SubscriptionClient


def get_default_subscription_id() -> str:
    """Get the default Azure subscription ID.

    Returns:
        str: The default Azure subscription ID.

    """
    try:
        azure_profile = json.loads(Path("~/.azure/azureProfile.json").expanduser().read_text())
    except json.JSONDecodeError:
        azure_profile = json.loads(Path("~/.azure/azureProfile.json").expanduser().read_text(encoding="utf-8-sig"))
    except FileNotFoundError:
        raise FileNotFoundError("Azure profile file not found. Please log in to Azure CLI.") from None

    return next(sub for sub in azure_profile["subscriptions"] if sub["isDefault"])["id"]


def get_subscription_username(subscription_id: str) -> str:
    """Get the username of the Azure account associated with the given subscription ID.

    Args:
        subscription_id (str): The Azure subscription ID.

    Returns:
        str: The username associated with the subscription ID.

    """
    try:
        azure_profile = json.loads(Path("~/.azure/azureProfile.json").expanduser().read_text())
    except json.JSONDecodeError:
        azure_profile = json.loads(Path("~/.azure/azureProfile.json").expanduser().read_text(encoding="utf-8-sig"))
    except FileNotFoundError:
        raise FileNotFoundError("Azure profile file not found. Please log in to Azure CLI.") from None

    return next(sub for sub in azure_profile["subscriptions"] if sub["id"] == subscription_id)["user"]["name"]


subscription_id = get_default_subscription_id()
username = get_subscription_username(subscription_id)
credential = DefaultAzureCredential()
client_authorization = AuthorizationManagementClient(credential, subscription_id)
client_registry = ContainerRegistryManagementClient(credential, subscription_id)
client_container = ContainerServiceClient(credential, subscription_id)
client_subscription = SubscriptionClient(credential)
client_network = NetworkManagementClient(credential, subscription_id)
client_resource = ResourceManagementClient(credential, subscription_id)


def get_location_availability_zones(location: str) -> list[str]:
    """Get the availability zones for a given Azure location.

    Args:
        location (str): Azure region to check for availability zones.

    Returns:
        list[str]: List of availability zones in the specified location.

    """
    loc = next(
        (loc for loc in client_subscription.subscriptions.list_locations(subscription_id) if loc.name == location),
        None,
    )
    return [zone.logical_zone for zone in (loc.availability_zone_mappings or [])]



"""Functions to manage Azure Kubernetes Service."""

from time import sleep
from uuid import uuid4

from azure.core.exceptions import HttpResponseError
from rich.progress import Progress, TaskID

from kina.azure import (
    client_authorization,
    client_container,
    client_network,
    get_location_availability_zones,
    subscription_id,
)


def get_latest_aks_version(location: str) -> str:
    """Get the available Kubernetes versions for AKS clusters.

    Args:
        credential (DefaultAzureCredential): Azure credentials for authentication.
        subscription_id (str): Azure subscription ID.
        location (str): Azure region to check for available Kubernetes versions.

    Returns:
        list[str]: List of available Kubernetes versions.

    """
    versions = client_container.managed_clusters.list_kubernetes_versions(location)
    return max(
        (patch for v in versions.values for patch in v.patch_versions),
        key=lambda ver: tuple(map(int, ver.split("."))),
    )


def create_aks_clusters(
    virtual_network_names: list[str],
    resource_group_name: str,
    rich_progress: Progress | None = None,
    rich_task: TaskID | None = None,
) -> list[str]:
    """Create Azure Kubernetes Service (AKS) clusters in the specified virtual networks.

    Args:
        virtual_network_names (list[str]): List of virtual network names to create AKS clusters in.
        resource_group_name (str): Name of the resource group to create the AKS clusters in.
        rich_progress (Progress | None, optional): Optional Rich Progress instance for progress tracking. Defaults to None.
        rich_task (TaskID | None, optional): Optional task ID for progress tracking. Defaults to None.

    Returns:
        list[str]: List of AKS cluster names created.

    """  # noqa: E501
    clusters = []

    for virtual_network_name in virtual_network_names:
        location = client_network.virtual_networks.get(resource_group_name, virtual_network_name).location
        vnet = client_network.virtual_networks.get(resource_group_name, virtual_network_name)
        subnet = client_network.subnets.get(resource_group_name, virtual_network_name, "kube_nodes")
        cluster_name = f"{resource_group_name}-{location}"
        if rich_progress and rich_task is not None:
            rich_progress.update(rich_task, description=f"Creating AKS Cluster: {cluster_name}", advance=1)
        client_container.managed_clusters.begin_create_or_update(
            resource_group_name=resource_group_name,
            resource_name=cluster_name,
            parameters={
                "location": location,
                "dns_prefix": cluster_name,
                "identity": {"type": "SystemAssigned"},
                "nodeResourceGroup": f"{resource_group_name}-{location}-nodes",
                "kubernetesVersion": get_latest_aks_version(location),
                "agent_pool_profiles": [
                    {
                        "name": "default",
                        "mode": "System",
                        "minCount": 2,
                        "maxCount": 6,
                        "enableAutoScaling": True,
                        "vm_size": "Standard_D2ads_v5",
                        "osSKU": "AzureLinux",
                        "vnet_subnet_id": subnet.id,
                        "osDiskSizeGB": 64,
                        "osDiskType": "Ephemeral",
                        "availabilityZones": get_location_availability_zones(location),
                    },
                ],
                "securityProfile": {"imageCleaner": {"enabled": True, "intervalHours": 24}},
                "oidcIssuerProfile": {"enabled": True},
                "networkProfile": {
                    "networkPlugin": "azure",
                    "networkPluginMode": "overlay",
                    "networkDataplane": "cilium",
                    "networkMode": "transparent",
                    "loadBalancerSku": "standard",
                    "podCidr": vnet.tags.get("pod-network"),
                    "serviceCidr": vnet.tags.get("service-network"),
                    "dnsServiceIP": vnet.tags.get("dns-service-ip"),
                    "ipFamilies": ["IPv4"],
                },
            },
        )
        clusters.append(cluster_name)
    return clusters


def configure_cluster_iam(
    resource_group: str,
    cluster_name: str,
) -> None:
    """Configure network IAM for the AKS cluster.

    Args:
        resource_group (str): Name of the resource group containing the AKS cluster.
        cluster_name (str): Name of the AKS cluster.

    Returns:
        None

    """
    role_ids = {
        "AcrPull": "7f951dda-4ed3-4680-a7ca-43fe172d538d",
        "NetworkContributor": "4d97b98b-1d4f-4787-a291-c67834d212e7",
    }
    for _ in range(30):
        try:
            cluster = client_container.managed_clusters.get(resource_group, cluster_name)
            principal_id = cluster.as_dict()["identity"]["principal_id"]
            kubelet_identity = cluster.as_dict()["identity_profile"]["kubeletidentity"]["object_id"]
            break
        except (KeyError, HttpResponseError):
            sleep(20)
            continue

    for role_id in role_ids.values():
        for identity in [principal_id, kubelet_identity]:
            for _ in range(30):
                try:
                    client_authorization.role_assignments.create(
                        scope=f"/subscriptions/{subscription_id}/resourceGroups/{resource_group}",
                        role_assignment_name=str(uuid4()),
                        parameters={
                            "role_definition_id": f"/providers/Microsoft.Authorization/roleDefinitions/{role_id}",
                            "principal_id": identity,
                        },
                    )
                    break
                except HttpResponseError:
                    sleep(20)
                    continue


def run_aks_command(
    resource_group_name: str,
    cluster_name: str,
    command: str,
) -> None:
    """Run a command on the AKS cluster.

    Args:
        credential (DefaultAzureCredential): Azure credentials for authentication.
        subscription_id (str): Azure subscription ID.
        resource_group_name (str): Name of the resource group containing the AKS cluster.
        cluster_name (str): Name of the AKS cluster.
        command (str): Command to run on the AKS cluster.

    Returns:
        None

    """
    client_container.managed_clusters.begin_run_command(
        resource_group_name=resource_group_name,
        resource_name=cluster_name,
        request_payload={"command": command},
    ).result()



"""Flux Operator for Kubernetes."""

from kina.azure.kubernetes import run_aks_command


def install_flux_operator(
    resource_group_name: str,
    cluster_name: str,
) -> None:
    """Install the Flux Operator on the specified AKS cluster.

    Args:
        credential (DefaultAzureCredential): Azure credentials for authentication.
        subscription_id (str): Azure subscription ID.
        resource_group_name (str): Name of the resource group containing the AKS cluster.
        cluster_name (str): Name of the AKS cluster to install the Flux Operator on.

    Returns:
        None

    """
    run_aks_command(
        resource_group_name,
        cluster_name,
        "helm install flux-operator oci://ghcr.io/controlplaneio-fluxcd/charts/flux-operator --namespace flux-system --create-namespace",  # noqa: E501
    )



"""Azure Container Registry (ACR) functions."""

from azure.mgmt.containerregistry.models import Registry, Sku

from kina.azure import client_registry


def create_acr(
    resource_group_name: str,
    location: str,
) -> str:
    """Create an Azure Container Registry (ACR) instance.

    Args:
        resource_group_name (str): Resource group name for the ACR instance.
        location (str): Azure region for the ACR instance.

    Returns:
        None

    """
    acr_name = resource_group_name.replace("-", "")
    client_registry.registries.begin_create(
        resource_group_name,
        acr_name,
        Registry(
            sku=Sku(name="Basic"),
            location=location,
            admin_user_enabled=True,
        ),
    ).result()
    return f"{acr_name}.azurecr.io"



"""Initialization module for the Kina CLI."""

from typing import Annotated

import typer
from rich.console import Console
from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
from rich.table import Table

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False)
kubectl = typer.Typer(no_args_is_help=True, help="Manage local Kubectl Configuration")
app.add_typer(kubectl, name="kubectl")


@app.command(name="create")
def cluster_create(
    locations: Annotated[
        str,
        typer.Option(
            "--locations",
            "--location",
            "-l",
            help="Azure Locations to create Kubernetes Clusters in, comma seperated",
        ),
    ] = "uksouth",
) -> None:
    """Create a new Kina AKS Cluster.

    Args:
        locations (str): Azure Locations to create Kubernetes Clusters in, comma separated.
        no_cni (bool): Disable CNI for the AKS clusters. This is useful for testing purposes.

    """
    from kina.azure.kubernetes import configure_cluster_iam, create_aks_clusters
    from kina.azure.registry import create_acr
    from kina.azure.resource_groups import create_resource_group, update_resource_group_status
    from kina.azure.virtual_networks import create_virtual_networks
    from kina.kubectl import add_cluster_to_kubeconfig, set_current_context

    locations = locations.split(",")
    cluster_count = len(locations)

    with Progress(
        SpinnerColumn(),
        BarColumn(),
        TimeElapsedColumn(),
        TextColumn("[progress.description]{task.description}"),
    ) as progress:
        task_rg = progress.add_task("Creating Resource Group...", total=1)
        task_vnet = progress.add_task("Creating Virtual Networks...", total=cluster_count + 1)
        task_acr = progress.add_task("Creating Container Registry...", total=1)
        task_aks = progress.add_task("Creating AKS Clusters...", total=cluster_count + 1)
        task_iam = progress.add_task("Creating Cluster IAM...", total=cluster_count + 1)
        task_kubectl = progress.add_task("Adding AKS Cluster to Kubeconfig...", total=cluster_count + 1)

        rg_name = create_resource_group(locations=locations)
        progress.update(task_rg, description=f"Created Resource Group: {rg_name}", advance=1)
        vnets = create_virtual_networks(
            locations=locations,
            resource_group_name=rg_name,
            rich_progress=progress,
            rich_task=task_vnet,
        )
        progress.update(task_vnet, description=f"Created Virtual Networks: {', '.join(vnets)}", advance=1)
        registry = create_acr(
            resource_group_name=rg_name,
            location=locations[0],
        )
        progress.update(task_acr, description=f"Created Container Registry: {registry}", advance=1)
        clusters = create_aks_clusters(
            virtual_network_names=vnets,
            resource_group_name=rg_name,
            rich_progress=progress,
            rich_task=task_aks,
        )
        progress.update(task_aks, description=f"Created AKS Clusters: {', '.join(clusters)}", advance=1)
        for cluster in clusters:
            progress.update(task_iam, description=f"Creating Cluster IAM: {cluster}", advance=1)
            configure_cluster_iam(
                resource_group=rg_name,
                cluster_name=cluster,
            )
        progress.update(task_iam, description=f"Created Cluster IAM: {', '.join(clusters)}", advance=1)
        for cluster in clusters:
            progress.update(task_kubectl, description=f"Adding AKS Cluster to Kubeconfig: {cluster}", advance=1)
            add_cluster_to_kubeconfig(
                resource_group_name=rg_name,
                cluster_name=cluster,
            )
        set_current_context(clusters[0])
        progress.update(
            task_kubectl,
            description=f"Added AKS Clusters to Kubeconfig: {', '.join(clusters)}, set current context to {clusters[0]}",  # noqa: E501
            advance=1,
        )
        update_resource_group_status(
            resource_group_name=rg_name,
            status="Running",
        )


@app.command(name="delete")
def cluster_delete(name: Annotated[str, typer.Argument()]) -> None:
    """Delete an existing Kina AKS Cluster.

    Args:
        name (str): The name of the Kina AKS Cluster to delete.

    """
    from kina.azure.resource_groups import delete_resource_group, update_resource_group_status
    from kina.kubectl import remove_from_kubeconfig

    update_resource_group_status(name, "Deleting")
    deleted = delete_resource_group(name)
    if deleted:
        typer.echo(f"Deleted Resource Group: {name}")
    else:
        typer.echo(f"Resource Group '{name}' not found or not managed by Kina.")

    remove_from_kubeconfig(name)


@app.command(name="list")
def cluster_list(
    output: Annotated[
        str,
        typer.Option("--output", "-o", help="Output format. Allowed values: json, table, names"),
    ] = "table",
) -> None:
    """List all Kina AKS Clusters.

    Args:
        output (str): Output format. Allowed values: json, table, names.

    """
    from kina.azure.resource_groups import list_resource_groups

    resource_groups = list_resource_groups()
    console = Console()

    if output == "table":
        t = Table()
        t.add_column("Name")
        t.add_column("Location(s)")
        t.add_column("Created By")
        t.add_column("Created At")
        t.add_column("Status")

        for resource_group_name, locations, username, created_at, status in resource_groups:
            t.add_row(resource_group_name, locations, username, created_at, status)
        console.print(t)
    elif output == "json":
        import json

        resource_groups_json = [
            {
                "name": resource_group_name,
                "locations": locations,
                "created_by": username,
                "created_at": created_at,
                "status": status,
            }
            for resource_group_name, locations, username, created_at, status in resource_groups
        ]
        console.print(json.dumps(resource_groups_json, indent=2))
    elif output == "names":
        resource_group_names = [resource_group_name for resource_group_name, _, _ in resource_groups]
        console.print("\n".join(resource_group_names))


@kubectl.command()
def cleanup() -> None:
    """Clean up the kubectl configuration by removing all Kina instances."""
    from kina.kubectl import cleanup_kubeconfig

    cleanup_kubeconfig()
    typer.echo("Cleaned up kubectl configuration")


@kubectl.command()
def remove(name: str) -> None:
    """Remove a Kina instance from the kubectl configuration.

    Args:
        name (str): The name of the Kina instance to remove.

    """
    from kina.kubectl import remove_from_kubeconfig

    remove_from_kubeconfig(name)
    typer.echo(f"Removed Kina Instance from Kubeconfig: {name}")


@app.command(name="version")
def version() -> None:
    """Display the installed version of Kina."""
    from importlib.metadata import version

    import requests

    console = Console()
    current_version = version("kina")
    try:
        latest_version = requests.get("https://pypi.org/pypi/kina/json", timeout=2).json()["info"]["version"]
    except requests.RequestException:
        latest_version = None

    if latest_version is None or current_version == latest_version:
        console.print(f"Kina version: [green]{current_version}[/]")
    else:
        console.print(f"Kina version: [red]{current_version}[/]")
        if latest_version:
            console.print(f"Latest version: [green]{latest_version}[/]")
            console.print('Update available! Run [bold]"uv tool update kina"[/] to update to the latest version.')
        else:
            console.print("Unable to check for updates.")


if __name__ == "__main__":
    app()
